#include <ctype.h>
#include <string.h>
#include <stdio.h> /* For (v)snprintf */

#include "log.h"
#include "rng.h"
#include "user.h"

static struct
 {
  char name[16];
  
  #ifdef CLIENT
  Colour c;
  int16_t x, y;
  #endif
 } user_start[128], *user_last = user_start, *user_end = user_start+128;

#ifdef SERVER

static int userHasName(const char *name)
 {
  uint8_t id = 0;
  
  if(name) for(; id < user_last-user_start; ++id)
   if(*user_start[id].name && !strcasecmp_custom(user_start[id].name, name))
    return 1;
   
  return 0;
 }

static void userConstructValidName(char *outbuf, int outbuflen, const char *in)
 {
  int outpos = 0;
  char out[16];
  
  if(in)
   {
    const char *in_start = in;
    
    for(; *in && outpos < 15; ++in)
     if(in == in_start)
      {
       if(isalpha(*in))
        out[outpos++] = *in;
       else in_start = in+1;
      }
     else if((*in == '_' && in[-1] != '_') || isalnum(*in))
      out[outpos++] = *in;
     
    out[outpos] = 0;
   }
  
  if(!in || !*out)
   {
    for(outpos = 0; outpos < 5; ++outpos)
     out[outpos] = 'a'+(RNG()%26);
    
    out[outpos] = 0;
   }
  
  while(userHasName(out))
   {
    int i;
    
    if(outpos + 3 > 15)
     outpos = 15-3;
    
    for(i = 0; i < 3; ++i)
     out[outpos+i] = '0'+(RNG()%10);
    
    out[outpos+3] = 0;
   }
  
  strncpy(outbuf, out, outbuflen);
  outbuf[outbuflen-1] = 0;
 }

uint8_t userGenerateNew(const char *name)
 {
  uint8_t id = 0;
  
  for(; id < user_last-user_start; ++id)
   if(!*user_start[id].name)
    {
     userConstructValidName(user_start[id].name, sizeof user_start->name, name);
     return id+1;
    }
  
  if(user_last != user_end)
   {
    userConstructValidName(user_last->name, sizeof user_start->name, name);
    ++user_last;
    return user_last-user_start;
   }
  
  return 0;
 }

#endif

uint8_t userFind(const char *name)
 {
  uint8_t id = 0, result = 0;
  int len = strlen(name);
  
  if(name) for(; id < user_last-user_start; ++id)
   if(*user_start[id].name && !strncasecmp_custom(name, user_start[id].name, len))
    {
     if(result)
      return 0;
     else result = id+1;
    }
   
  return result;
 }

void userForgetAll(void)
 {
  user_last = user_start;
 }

void userForget(uint8_t id)
 {
  if(id && id-1 < user_last-user_start && *user_start[id-1].name)
   *user_start[id-1].name = 0;
 }

int userValid(uint8_t id)
 {
  return id && id-1 < user_last-user_start && *user_start[id-1].name;
 }

const char * userGetName(uint8_t id)
 {
  if(id && id-1 < user_last-user_start && *user_start[id-1].name)
   return user_start[id-1].name;
  else
   {
    logMessage(Debug, "Requested name of invalid user (%d)", id);
    return 0;
   }
 }

void userSetName(uint8_t id, const char *name)
 {
  if(name && id && id-1 < user_end - user_start)
   {
    while(user_last < user_start + id)
     {
      *user_last->name = 0;
      ++user_last;
     }
    
    strncpy(user_start[id-1].name, name, sizeof user_start->name-1);
    user_start[id-1].name[sizeof user_start->name-1] = 0;
   }
 }

#ifdef CLIENT

#ifdef USE_ALFONT
#include <alfont.h>
extern ALFONT_FONT *alfont;
#endif

#include "config.h"
#include "region.h"

extern BITMAP *buffer;
extern DirtyRectangles dirt;
extern int shadowColour(Colour c);

static void redrawUser(uint8_t id)
 {
  if(id && id-1 < user_last-user_start && *user_start[id-1].name)
   {
    int length, height, shadow = shadowColour(user_start[id-1].c);
    char buf[64];
    Region region;
    
    snprintf(buf, sizeof buf, "%d:%s", id, user_start[id-1].name);
    
    #ifdef USE_ALFONT
    if(alfont)
     {
      length = alfont_text_length(alfont, buf);
      height = alfont_text_height(alfont);
     }
    else
    #endif
     {
      length = text_length(font, buf);
      height = text_height(font);
     }
    
    region.x1 = user_start[id-1].x-length/2+(screen_w-buffer_w)/2-1;
    region.y1 = user_start[id-1].y-height/2+(screen_h-buffer_h)/2-1;
    region.x2 = region.x1+length+2;
    region.y2 = region.y1+height+2;
    
    #ifdef USE_ALFONT
    if(alfont)
     {
      alfont_textout_aa_ex(buffer, alfont, buf, region.x1,   region.y1+1, shadow, -1);
      alfont_textout_aa_ex(buffer, alfont, buf, region.x1+1, region.y1,   shadow, -1);
      alfont_textout_aa_ex(buffer, alfont, buf, region.x1+2, region.y1+1, shadow, -1);
      alfont_textout_aa_ex(buffer, alfont, buf, region.x1+1, region.y1+2, shadow, -1);
      alfont_textout_aa_ex(buffer, alfont, buf, region.x1+1, region.y1+1, makecol32(user_start[id-1].c.r, user_start[id-1].c.g, user_start[id-1].c.b), -1);
     }
    else
    #endif
     {
      #ifdef OLD_ALLEGRO
      textout(buffer, font, buf, region.x1,   region.y1+1, shadow);
      textout(buffer, font, buf, region.x1+1, region.y1,   shadow);
      textout(buffer, font, buf, region.x1+2, region.y1+1, shadow);
      textout(buffer, font, buf, region.x1+1, region.y1+2, shadow);
      textout(buffer, font, buf, region.x1+1, region.y1+1, makecol32(user_start[id-1].c.r, user_start[id-1].c.g, user_start[id-1].c.b));
      #else
      textout_ex(buffer, font, buf, region.x1,   region.y1+1, shadow, -1);
      textout_ex(buffer, font, buf, region.x1+1, region.y1,   shadow, -1);
      textout_ex(buffer, font, buf, region.x1+2, region.y1+1, shadow, -1);
      textout_ex(buffer, font, buf, region.x1+1, region.y1+2, shadow, -1);
      textout_ex(buffer, font, buf, region.x1+1, region.y1+1, makecol32(user_start[id-1].c.r, user_start[id-1].c.g, user_start[id-1].c.b), -1);
      #endif
     }
    
    DRAdd(&dirt, &region);
   }
 }

/*static void undrawUser(uint8_t id)
 {
  if(id && id-1 < user_last-user_start && *user_start[id-1].name)
   {
    Region region;
   }
 }*/

void redrawAllUsers(void)
 {
  uint8_t id = 0;
  for(; id < user_last-user_start; ++id) if(*user_start[id].name)
   redrawUser(id+1);
 }

void userSetColour(uint8_t id, Colour c)
 {
  if(id && id-1 < user_last-user_start && *user_start[id-1].name)
   {
    user_start[id-1].c = c;
   }
  else logMessage(Debug, "Requested set colour of invalid user (%d)", id);
 }

void userSetPos(uint8_t id, int16_t x, int16_t y)
 {
  if(id && id-1 < user_last-user_start && *user_start[id-1].name)
   {
    user_start[id-1].x = x;
    user_start[id-1].y = y;
   }
  else logMessage(Debug, "Requested set position of invalid user (%d)", id);
 }
/*
void userGetColour(uint8_t id, Colour *c)
 {
  if(id && id-1 < user_last-user_start && *user_start[id-1].name)
   *c = user_start[id-1].c;
 }

void userGetPos(uint8_t id, int16_t *x, int16_t *y)
 {
  if(id && id-1 < user_last-user_start && *user_start[id-1].name)
   {
    *y = user_start[id-1].x;
    *x = user_start[id-1].y;
   }
 }
*/
#endif
