#include "videodump.h"
#include "mem.h"

#include <assert.h>

static ScaleInfo *buildScaleInfo(unsigned int original, unsigned int dest)
 {
  unsigned int p = 0;
  
  ScaleInfo *info = (ScaleInfo *)malloc(sizeof(ScaleInfo)*dest);
  if(!info) return 0;
  
  if(dest > original)
   {
    // Special case: Enlarging. Simple linear interpolation.
    ScaleInfo *i = info;
    
    for(; p < dest; ++p)
     {
      i->end = (i->start = (float *)malloc(sizeof(float)*2))+2;
      i->base = p*original/dest;
      
      *i->start = 1.0f - (i->start[1] = p*original/(float)(dest-1)-p*original/(dest-1));
      
      if(*i->start > 0.99999f)
       {
        *i->start = 1;
        i->end = i->start+1;
       }
      else if(*i->start < 0.00001f)
       {
        *i->start = 1;
        i->end = i->start+1;
        ++i->base;
       }
      
      ++i;
     }
    
    return info;
   }
  
  for(; p < dest; ++p)
   {
    float start = p*original/(float)dest, end = (p+1)*original/(float)dest;
    unsigned int istart, iend;
    info[p].start = (float *)malloc(sizeof(float)*((original+dest-1)/dest+1));
    info[p].base = istart = (unsigned int)start;
    iend = (unsigned int)end;
    
    if(istart == iend)
     {
      info[p].end = info->start+1;
      *info[p].start = 1.0f;
     }
    else
     {
      float *out = info[p].start, sum;
      sum = *out = 1.0f-(start-istart);
      
      for(++out, ++istart; istart != iend; ++out, ++istart)
       sum += (*out = 1.0f);
      
      sum += (*out = end-iend);
      if(*out > 0.00001f && iend != original) ++out;
      
      info[p].end = out;
      
      sum = 1.0f/sum;
      
      for(out = info[p].start; out != info[p].end; ++out)
       *out *= sum;
     }
   }
  
  return info;
 }

static void destroyScaleInfo(ScaleInfo *info, unsigned int length)
 {
  unsigned int p = 0;
  for(; p < length; ++p) free(info[p].start);
  free(info);
 }

VideoDump *createVideoDump(FILE *out, BITMAP *in, unsigned int width, unsigned int height)
 {
  VideoDump *dump;
  
  /* Width and height need to be positive even numbers. */
  assert(width > 0 && !(width&1));
  assert(height > 0 && !(height&1));
  assert(in);
  assert(out);
  assert(bitmap_color_depth(in) == 32);
  
  if(!(dump = (VideoDump *)malloc(sizeof(VideoDump))))
   return 0;
  
  dump->w = width;
  dump->h = height;
  dump->out = out;
  dump->in = in;
  
  if(!(dump->frame_buffer = (unsigned char *)malloc(width*height*3/2)))
   {
    free(dump);
    return 0;
   }
  
  if(!(dump->x = buildScaleInfo(in->w, width)))
   {
    free(dump->frame_buffer);
    free(dump);
    return 0;
   }
  
  if(!(dump->y = buildScaleInfo(in->h, height)))
   {
    destroyScaleInfo(dump->x, width);
    free(dump->frame_buffer);
    free(dump);
    return 0;
   }
  
  return dump;
 }

void releaseVideoDump(VideoDump *dump)
 {
  destroyScaleInfo(dump->y, dump->h);
  destroyScaleInfo(dump->x, dump->w);
  free(dump->frame_buffer);
  free(dump);
 }

int performDump(VideoDump *dump)
 {
  unsigned int w=dump->w, h=dump->h;
  unsigned char *outy = dump->frame_buffer,
                *outu = outy+w*h,
                *outv = outu+(w*h/4);
  
  BITMAP *bmp = dump->in;
  const ScaleInfo *_xs=dump->x, *ys=dump->y, *xse=_xs+w, *yse=ys+h;
  unsigned char even=3;
  int p;
  float s;
  
  for(; ys != yse; ++ys)
   {
    float u = 0, v = 0;
    
    const ScaleInfo *xs = _xs;
    
    for(; xs != xse; ++xs)
     {
      float r = 0, g = 0, b = 0;
      
      unsigned int ix, _ix = xs->base;
      unsigned int iy = ys->base;
      const float *_xw = xs->start, *xw,  *xwe = xs->end, *yw = ys->start, *ywe = ys->end;
      
      
      for(; yw != ywe; ++yw, ++iy)
       {
        ix = _ix;
        xw = _xw;
        
        for(; xw != xwe; ++xw, ++ix)
         {
          assert(ix < (unsigned int)bmp->w);
          assert(iy < (unsigned int)bmp->h);
          p = _getpixel32(bmp, ix, iy);
          s = *xw**yw;
          r += getr32(p)*s;
          g += getg32(p)*s;
          b += getb32(p)*s;
         }
       }
      
      *outy = (unsigned char)(0.257f*r+0.504f*g+0.098f*b+16.0f); ++outy;
      
      u += (-0.148f)*r+(-0.291f)*g+0.439f*b+128.0f;
      v += 0.439f*r+(-0.368f)*g+(-0.071f)*b+128.0f;
      
      if((even ^= 2) & 2)
       {
        if(even & 1)
         {
          *outu = (unsigned char)(u*0.5f); ++outu;
          *outv = (unsigned char)(v*0.5f); ++outv;
         }
        else
         {
          *outu = (unsigned char)(*outu*0.5f+u*0.25f); ++outu;
          *outv = (unsigned char)(*outv*0.5f+v*0.25f); ++outv;
         }
        
        u = 0.0f; v = 0.0f;
       }
     }
    
    if(!((even ^= 1)&1))
     {
      outu -= w/2;
      outv -= w/2;
     }
   }
  
  return 1 == fwrite(dump->frame_buffer, 3*w*h/2, 1, dump->out);
 }

int duplicatePreviousFrame(VideoDump *dump)
 {
  return 1 == fwrite(dump->frame_buffer, 3*dump->w*dump->h/2, 1, dump->out);
 }
